# Pfilter/Expand.pm
#
# common library functions for the macro/define/constant expanding parts of pfilter

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# list of functions in this file:

# function to evaluate a conditional string
# sub evaluate_conditional( $vars_ref, $conditional, $error_lines_ref, 
# 			    $warning_lines_ref, $syntax_error_ref,
# 			    $evaluated_ref )

# function to expand/consume a single conditional block that starts at
# sub expand_conditional_block_in_lines_at_offset( $vars_ref, $lines_ref, 
# 						   $conditional_start_index,
#                                                  $fields_without_comment_ref,
# 						   $error_lines_ref, 
# 						   $warning_lines_ref )

# function to expand/replace the constants/defines in a single line
# sub expand_constants_defines_in_lines_at_offset( $vars_ref, $lines_ref, 
# 						   $line_index, 
# 						   $constants_ref, 
# 						   $defines_ref,
# 						   $error_lines_ref, 
# 						   $warning_lines_ref )

# function to expand/replace the constants/defines in a list of lines
# sub expand_constants_defines_in_lines( $vars_ref, $lines_ref, 
# 				         $constants_ref, $defines_ref,
# 				         $error_lines_ref,
# 				         $warning_lines_ref )
 
# function to expand/replace the constants/defines in a string
# sub expand_constants_defines_in_string( $vars_ref, $string_ref, 
# 					  $constants_ref, $defines_ref,
# 					  $error_lines_ref
# 					  $warning_lines_ref )

# function to expand/replace .. ranges in an array of strings
# sub expand_dot_dot_ranges_in_strings( $vars_ref,
#			                $strings_ref,
#                                       $error_lines_ref,
#			                $warning_lines_ref )

# function to expand a loop in a list of text lines at a given offset
# sub expand_loop_in_lines_at_offset( $vars_ref, $lines_ref, 
#                                     $loop_start_index,
#                                     $fields_without_comment_ref,
#                                     $error_lines_ref,
#                                     $warning_lines_ref )

# function to expand the macros etc in a list of text lines
# sub expand_lines( $vars_ref, $lines_ref, $error_lines_ref, $warning_lines_ref )

# function to expand a macro in a list of text lines at a given offset
# sub expand_macro_in_lines_at_offset( $vars_ref, $lines_ref, $macro_start_index,
# 				       $error_lines_ref, $warning_lines_ref )

# function to expand/process/consume a metadirective in a list of text 
# sub expand_metadirective_in_lines_at_offset( $vars_ref, $lines_ref, 
# 					       $line_index,
#                                              $fields_without_comment_ref,
# 					       $error_lines_ref,
# 					       $warning_lines_ref )

# function to expand/consume a single trailing conditional in a line at
# sub expand_trailing_conditional_in_lines_at_offset( $vars_ref, $lines_ref,
# 						      $conditional_start_index,
#                                                     $fields_without_comment_ref,
# 						      $error_lines_ref,
# 						      $warning_lines_ref )

#=======================================================================

# function to evaluate a conditional string
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   conditional string
#   reference to error lines array
#   reference to warning lines array
#   reference to syntax error flag
#   reference to conditional evaluated true flag
# This function modifies the referenced flags.
# Note that all referenced defines/constants in the conditional 
# arguments are substituted with their values before evaluation.

sub evaluate_conditional
{
    my( $vars_ref, $conditional, $error_lines_ref, $warning_lines_ref,
	$syntax_error_ref, $evaluated_ref ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};
#    print "$0: in evaluate_conditional conditional=<$conditional>\n"
#	if $debug;

    # make sure we were called with a valid conditional
    if ( $conditional !~ /^\s*\%(if|ifdef|ifndef)(\s+|$)/ ) {
	push @$error_lines_ref, "internal error - trying to evaluate unknown type of conditional <$conditional> in subroutine evaluate_conditional";
	$$syntax_error_ref = 1;
	return;
    }

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};

    # find the conditional without any trailing comment, and
    # find the conditional expression/arguments
    my $wo_comment = join( " ", 
			   fields_without_ruleset_comment( $vars_ref,
							   $conditional ) );
    my $arguments = $wo_comment;
    $arguments =~ s/^[^\s]+([\s]+|$)//;
#    print "$0: in evaluate_conditional wo_comment=<$wo_comment> arguments=<$arguments>\n"
#	if $debug;

    # do any define/constant substitutions in the 
    # condtional expression/arguments

    my $expanded_wo_comment = $wo_comment;
    expand_constants_defines_in_string( $vars_ref,
					\$expanded_wo_comment,
					$$vars_ref{constants},
					$$vars_ref{defines},
					$error_lines_ref,
					$warning_lines_ref );

    # split the non-comment part of the conditional expression
    # into non-white-space fields, and seperate out the type of
    # conditional, and an array of the arguments
    my @expanded_fields = split(' ',$expanded_wo_comment);
    my $type = shift @expanded_fields;
    my $expanded_arguments = join( " ", @expanded_fields );
#    if ( $debug ) {
#	print "$0: in evaluate_conditional wo_comment=<$wo_comment> expanded_wo_comment=<$expanded_wo_comment>\n";
#	print "$0: in expanded_fields=<@expanded_fields> evaluate_conditional type=<$type> expanded_arguments=<$expanded_arguments>\n";
#    }

    # take care of the easy ones first, the %ifdef and %ifndef
    if ( $type eq "%ifdef" || $type eq "%ifndef" ) {
	if ( ! @expanded_fields ) {
	    push @$error_lines_ref, "conditional $type is missing the constant/define argument";
	    $$syntax_error_ref = 1;
	    return;
	}
	# my $name = shift @fields;
	my $name = shift @expanded_fields;
	if ( @expanded_fields ) {
	    push @$error_lines_ref, "conditional <$wo_comment> has too many arguments";
	    $$syntax_error_ref = 1;
	    return;
	}
        if ( $name !~ /^[A-Za-z\_][A-Za-z0-9\_\-]*$/ ) {
	    push @$error_lines_ref, "conditional $type has bad constant/define name <$name>";
	    $$syntax_error_ref = 1;
	    return;
	}
	my $found = ( exists $$constants_ref{$name} || $$defines_ref{$name} );
	$$evaluated_ref = ( $type eq "%ifdef" ) ? $found : ! $found;
        $$syntax_error_ref = 0;
#	print "$0: in evaluate_conditional conditional=<$conditional> evaluated $$evaluated_ref\n"
#	    if $debug;
	return;
    }

    # now for the hard one, the %if
    if ( $type eq "%if" ) {

        # the easy case is %if with no arguments
        # which means something evaluated to blank
        # and we return false
        if ( ! @expanded_fields ) {
	    $$syntax_error_ref = 0;
	    $$evaluated_ref = 0;
#	    print "$0: in evaluate_conditional conditional=<$conditional> evaluated $$evaluated_ref\n"
#		if $debug;
            return;
        }

        # try for the "%if -T FILE" cases now
        if ( $expanded_fields[0] =~ /^\-/ ) {
            my $test = shift @fieilds;
	    if ( ! @fields ) {
	        push @$error_lines_ref, "conditional <$expanded_wo_comment> is missing the file/directory argument";
	        $$syntax_error_ref = 1;
	        return;
	    }
            my $name = shift @fields;
	    if ( @expanded_fields ) {
	        push @$error_lines_ref, "conditional <$expanded_wo_comment> has too many arguments";
	        $$syntax_error_ref = 1;
	        return;
	    }
            if ( $test eq "-r" ) {
		$$evaluated_ref = ( -r "$name" );
		$$syntax_error_ref = 0;
	        return;
            } elsif ( $test eq "-w" ) {
		$$evaluated_ref = ( -w "$name" );
		$$syntax_error_ref = 0;
	        return;
            } elsif ( $test eq "-x" ) {
		$$evaluated_ref = ( -x "$name" );
		$$syntax_error_ref = 0;
	        return;
            } elsif ( $test eq "-d" ) {
		$$evaluated_ref = ( -d "$name" );
		$$syntax_error_ref = 0;
	        return;
	    } else {
	        push @$error_lines_ref,
		"conditional <$wo_comment> has unknown test <$test>";
	        $$syntax_error_ref = 1;
	        return;
	    }
        }

        # another easy one is "%if string" with no "=" characters in the string
        if ( $expanded_arguments !~ "\=" ) {
            $$evaluated_ref = ( $expanded_arguments ne "" &&
				$expanded_arguments ne "0" ) ?
				    1 : 0;
	    $$syntax_error_ref = 0;
	    return;
	}

        # now for the = and != conditional tests
        my @strings_wo_equal_chars = split('\=', $expanded_arguments);
        if ( scalar @strings_wo_equal_chars > 2 ) {
	    push @$error_lines_ref,
	    "conditional <$expanded_wo_comment> cannot have more than one = seperator";
	    $$syntax_error_ref = 1;
	    return;
	}
        # if we're doing != ...
        if ( $strings_wo_equal_chars[0] =~ /\!$/ ) {   
            $strings_wo_equal_chars[0] =~ s/\!$//;
	    $strings_wo_equal_chars[0] =~ s/\s*$//;
	    $strings_wo_equal_chars[1] =~ s/^\s*//;
            $$evaluated_ref =
		( $strings_wo_equal_chars[0] ne $strings_wo_equal_chars[1] );
	    $$syntax_error_ref = 0;
	    return;
        # or if we're doing = ...
        } else {
	    $strings_wo_equal_chars[0] =~ s/\s*$//;
	    $strings_wo_equal_chars[1] =~ s/^\s*//;
            $$evaluated_ref = 
		( $strings_wo_equal_chars[0] eq $strings_wo_equal_chars[1] );
	    $$syntax_error_ref = 0;
	    return;
        }
    }

    # unknown conditional type
    push @$error_lines_ref, "internal error - cannot evaluate unknown type of conditional <$expanded_wo_comment> in subroutine evaluate_conditional";
    $$syntax_error_ref = 1;
    return;
}

#=======================================================================

# function to expand/consume a single conditional block that starts at
# a given offset in a list of text lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   offset into the lines array that has the starting %if? line
#   reference to array of non-blank fields excluding ruleset comments
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines, deleting the 
# starting %if? line and the matching %endif line, and the lines
# in between if the conditional did not evaluate true.
# Returned parameters are:
#   This function returns non-zero if a conditional was processed
# Note that all referenced defines/constants in the conditional 
# arguments are substituted with their values before evaluation.

sub expand_conditional_block_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $conditional_start_index,
	$fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};

#    if ( $debug ) {
#	my $string = ( $conditional_start_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$conditional_start_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_conditional_block_in_lines_at_offset offset=$conditional_start_index $string";
#    }

    return 0 if ( $conditional_start_index > $#$lines_ref );

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};

    # get the line pointed to by the offset
    my $line_index = $conditional_start_index;
    my $line = $$lines_ref[$line_index];

    # see if the line starts with a conditional expression.
    # if so, we need to identify the complete conditional
    # block and process it, skipping over nested blocks 
    if ( @$fields_without_comment_ref &&
	 $$fields_without_comment_ref[0] =~ /^\%(if|ifdef|ifndef)$/ ) {
	    
	# find the conditional expression without
	# any trailing comment
	my $if_index = $line_index;
	my $conditional = join( " ", @$fields_without_comment_ref );
	    
	# find the end of the conditional block making sure to
	# skip over nested embedded conditional blocks,
	# saving the location of the %else if it exists
	if ($debug ) {
	    print "$0: in expand_conditional_block_in_lines_at_offset block starting with:\n";
	    printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
	}
	my $conditional_level = 1;          # start at level 1 because 1 %if? found
	my @else_indexes = ();              # empty means no %else found yet
	my $endif_index = -1;               # -1 means no %endif found yet
	my $search_index = $line_index + 1; # start at the line after the %if?
	# loop through lines until %endif found or no more lines
	while ( $endif_index == -1
		&& $search_index <= $#$lines_ref ) {
	    my $search_line = $$lines_ref[$search_index];
	    if ( $search_line =~ /^[\s]*\%(if|ifdef|ifndef)(\s|$)/ ) {
		$conditional_level += 1;
	    } elsif ( $search_line =~ /^\s*\%else(\s|$)/
					       && $conditional_level <= 1 ) {
		push @else_indexes, $search_index;
#		print "$0: in expand_conditional_block_in_lines_at_offset after adding else $search_index elses=@else_indexes\n"
#		    if $debug;
	    } elsif ( $search_line =~ /^\s*\%endif(\s|$)/ ) {
		$conditional_level -= 1;
		$endif_index = $search_index
		    if ( ! $conditional_level );
	    }
	    # go to next line
	    $search_index += 1;
	}
#	print "$0: in expand_conditional_block_in_lines_at_offset after searching level=$conditional_level elses=@else_indexes endif=$endif_index search=$search_index last=$#$lines_ref\n"
#	    if $debug;

	# evaluate the conditional expression
	my $syntax_error = 0;
	my $evaluated = 0;
	evaluate_conditional( $vars_ref, $conditional,
			      $error_lines_ref, $warning_lines_ref,
			      \$syntax_error, \$evaluated );
#	print "$0: in expand_conditional_block_in_lines_at_offset after evaluating <$conditional> syntax_error=$syntax_error evaluated=$evaluated\n"
#	    if $debug;

        # if we cannot find the end of the conditional block,
	# note the error
	push @$error_lines_ref, "conditional <$conditional> has no matching %endif"
	    if ( $endif_index == -1 );

	# get rid of the line containing the initial %if?, this
        # also means we have to decrement the indexes we found
        # for any appropriate %else lines and the %endif line
	splice @$lines_ref, $if_index, 1;
        foreach my $else_index ( @else_indexes ) {
	    $else_index--;
	}
	$endif_index--
	    if ( $endif_index != -1 );
#	print "$0: in expand_conditional_block_in_lines_at_offset after munching %if line line_index=$line_index elses=@else_indexes endif=$endif_index last=$#$lines_ref\n"
#	    if $debug;

	# now for every block up to an %else,
        # and the block up to the %endif,
        # we include the block text if there has been a syntax
        # error or if the conditional was evaluated to true. 
        # we complement the evaluated value after every 
        # block so that alternating blocks around %else
        # statements are included/not included
        my @block_indexes = @else_indexes;
        push @block_indexes, $endif_index;
	while ( @block_indexes ) {
#	    print "$0: in expand_conditional_block_in_lines_at_offset block line_index=$line_index block_indexes=@block_indexes last_index=$#$lines_ref evaluated=$evaluated\n"
#		if $debug;
	    if ( $syntax_error || $evaluated ) {
		if ( $debug ) {
		    my $debug_index = $line_index;
		    my $debug_count = $block_indexes[0] - $line_index;
		    print "$0: in expand_conditional_block_in_lines_at_offset keeping $debug_count lines:\n";
		    while ( $debug_index <= $#$lines_ref &&
			    $debug_count > 0) {
			printf "%s: line %04d: <%s>\n", $0, $debug_index, $$lines_ref[$debug_index];
			$debug_index++;
			$debug_count--;
		    }
		}
		# since we are including this text, bump our
		# current line index past the text up to where
		# the %else is
		$line_index = $block_indexes[0];
	    } else {
		# figure out how many text lines to get rid of
		my $count = $block_indexes[0] - $line_index;
		if ( $count < 0 ) {
		    push @$error_lines_ref, "internal error - negative count before deleting pre %else block for conditional <$conditional>";
		    $count = 0;
		}
		if ( $debug ) {
		    my $debug_index = $line_index;
		    my $debug_count = $count;
		    print "$0: in expand_conditional_block_in_lines_at_offset deleting $count lines:\n";
		    while ( $debug_index <= $#$lines_ref &&
			    $debug_count > 0 ) {
			printf "%s: line %04d: <%s>\n", $0, $debug_index, $$lines_ref[$debug_index];
			$debug_index++;
			$debug_count--;
		    }
		}
		# get rid of the text lines
		splice @$lines_ref, $line_index, $count;
		# adjust all the remaining block indexes
		foreach my $block_index ( @block_indexes ) {
		    $block_index -= $count;
		}
#		print "$0: in expand_conditional_block_in_lines_at_offset after deleting/adjusting line_index=$line_index block_indexess=@block_indexes last=$#$lines_ref\n"
#		    if $debug;
	    }
	    # now get rid of this %else/%endif from the text
	    my $block_index = shift @block_indexes; # get index of %else/%endif
	    splice @$lines_ref, $line_index, 1;
	    foreach my $block_index ( @block_indexes ) {
		$block_index--;
	    }
#	    print "$0: in expand_conditional_block_in_lines_at_offset after deleting %else/%endif line line_index=$line_index block_indexes=@block_indexes last=$#$lines_ref\n"
#		if $debug;
	    # reverse the evaluated value since we just passed a %else
	    $evaluated = ( ! $evaluated );		    
	}

        return 1;		      

    } else {

	return 0;

    }
}

#=======================================================================

# function to expand/replace the constants/defines in a single line
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to list of lines
#   offset of line to expand in list of lines
#   reference to hash of constant/value pairs
#   reference to hash of define/value pairs
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_constants_defines_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $line_index, $constants_ref, $defines_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;

#    if ( ${$$vars_ref{constants}}{debug} ) {
#	my $string = ( $line_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$line_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_constants_defines_in_lines_at_offset offset=$line_index $string";
#    }

    # do the substituions
    my $line = $$lines_ref[$line_index];
    my $expanded = expand_constants_defines_in_string( $vars_ref,
						       \$line,
						       $constants_ref,
						       $defines_ref,
						       $error_lines_ref,
						       $warning_lines_ref );
    $$lines_ref[$line_index] = $line if $expanded;

    return $expanded;
}

#=======================================================================

# function to expand/replace the constants/defines in a list of lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   reference to hash of constant/value pairs
#   reference to hash of define/value pairs
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_constants_defines_in_lines
{
    my( $vars_ref, $lines_ref, $constants_ref, $defines_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};
    my $expanded = 0;

    if ( $debug ) {
	my $count = scalar @$lines_ref;
	print "$0: in expand_constants_defines_in_lines $count lines\n";
	print_hash( "", "in expand_constants_defines_in_lines constants hash", $constants_ref );
	print_hash( "", "in expand_constants_defines_in_lines defines hash", $defines_ref );
    }

    # loop through the input lines doing expansions on each
    my $line_index = 0;
    while ( $line_index <= $#$lines_ref ) {
        $expanded |= expand_constants_defines_in_lines_at_offset
            ( $vars_ref,
	      $lines_ref,
	      $line_index, 
	      $constants_ref,
	      $defines_ref,
	      $error_lines_ref, 
	      $warning_lines_ref );
	$line_index++;
    }

    return ( $expanded );
}

#=======================================================================

# function to expand/replace the constants/defines in a string
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to string to munge
#   reference to hash of constant/value pairs
#   reference to hash of define/value pairs
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_constants_defines_in_string
{
    my( $vars_ref, $string_ref, $constants_ref, $defines_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};
    my $expanded = 0;

    print "$0: in expand_constants_defines_in_string string=<$$string_ref>\n"
        if $debug;

    # if there is a possible %name% in the string, ...
    my $string = $$string_ref;
    if ( $string =~ /\%[A-Za-z\_][A-Za-z0-9\_\-]*\%/ ) {

	# split the line into fields by the % characters
	my $between = $string;
	$between =~ s/^[^\%]*\%/%/;
	$between =~ s/\%[^\%]*$/%/;
	my @fields = split('%', $between);
	
	# loop through the fields starting at the beginning,
	# for each one that matches a %constant or %define,
	# do the substition and dump the next field since
	# using the first fields in %used%next% means the next
	# field does not have a % at the beginning any more
	while ( @fields ) {
	    my $name = shift @fields;
	    if ( $name =~ /^[A-Za-z\_][A-Za-z0-9\_\-]*$/ ) {
		if ( exists $$constants_ref{$name} ) {
		    print "$0: substituting <$$constants_ref{$name}> for %$name%\n"
			if $debug;
		    $string =~ s/\%$name\%/$$constants_ref{$name}/;
		    shift @fields if ( @fields );
		    $expanded = 1;
		} elsif ( exists $$defines_ref{$name} ) {
		    print "$0: substituting <$$defines_ref{$name}> for %$name%\n"
			if $debug;
		    $string =~ s/\%$name\%/$$defines_ref{$name}/;
		    shift @fields if ( @fields );
		    $expanded = 1;
#		} else {
#		    print "$0: cannot find %constant or %define for %$name%\n"
#			if $debug;
		}
	    }
	}
	$$string_ref = $string if $expanded;
    }

    return ( $expanded );
}

#=======================================================================

# function to expand/replace .. ranges in an array of strings
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to strings array
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of strings
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_dot_dot_ranges_in_strings
{
    my( $vars_ref,
	$strings_ref, 
	$error_lines_ref,
	$warning_lines_ref ) = @_;

    # simplest case to be effecient

    return 0 if scalar @$strings_ref < 3;

    # if no .. strings in the middle of the array, done

    return 0 if 
	! grep( /^\.\.$/, @$strings_ref[ 1 .. ($#$strings_ref - 1) ] );

    my $debug = ${$$vars_ref{constants}}{debug};
    my $expanded = 0;

    print "$0: in expand_dot_dot_ranges_in_strings strings=<@$strings_ref>\n"
        if $debug;

    # loop through the strings but not the beginning or ending string

    my $strings_index = 0;
    while ( $strings_index <= ( $#$strings_ref - 2 ) ) {

	# if this string is a .. string

	if ( $$strings_ref[$strings_index + 1] =~ /^\.\.$/ ) {
	    
	    # examine the strings before and after it
	    my $string_before = $$strings_ref[$strings_index];
	    my $string_before_base   = $string_before;
	       $string_before_base   =~ s/[0-9]+$//;
	    my $string_before_number = $string_before;
	       $string_before_number =~ s/^.*[^0-9]//;
	    
	    my $string_after  = $$strings_ref[$strings_index + 2];
	    my $string_after_base   = $string_after;
	       $string_after_base   =~ s/[0-9]+$//;
	    my $string_after_number = $string_after;
	       $string_after_number =~ s/^.*[^0-9]//;

	    print "$0: in expand_dot_dot_ranges_in_strings before=$string_before base=$string_before_base number=$string_before_number after=$string_after base=$string_after_base number=$string_after_number\n" if $debug;

	    # make sure range is stringNUM1 .. stringNUM2

	    if ( $string_before_base ne 
		     $string_after_base ||
		 $string_before_number eq "" ||
		 $string_after_number eq "" ) {
		print "$0: in expand_dot_dot_ranges_in_strings illegal range\n"
		    if $debug;
		push @$error_lines_ref,
		"illegal range $string_before .. $string_after";
	    } else {

		# we have a winner, we need to swallow 
		# the range strings

		splice @$strings_ref, $strings_index, 3;
		$expanded = 1;

		# if the numbers are reversed, output a
		# warning message

		push @$warning_lines_ref,
		"range $string_before .. $string_after is reversed"
		    if $string_before_number > 
		       $string_after_number;

		# insert new strings for the range,
		# making sure that if the first number
		# string started with a 0 character and
		# it is the same length as the second
		# number string, all number strings
		# are padded at the beginning if needed

		my $format = ( length $string_before_number ==
			       length $string_after_number
			       && $string_before_number =~ /^0/ ) ?
			       "%0" . length($string_before_number) . "d" :
			       "%d";;
		my $number = $string_before_number;
		my $limiting = 0;
		my $limit = 256;
		while ( $number <= $string_after_number &&
			( $limiting < $limit ||
			  ! ${$$vars_ref{constants}}{limits} ) ) {
		    my $insertion = $string_before_base .
			sprintf( "$format", $number);
		    print "$0: in expand_dot_dot_ranges_in_strings inserting $insertion\n"
			if $debug;
		    splice @$strings_ref, $strings_index, 0, $insertion;
		    $strings_index++;
		    $number++;
		    $limiting++;

		} # end inserting range strings

	        # if we ran into an expansion limit,
	        # tell of it

	        push @$error_lines_ref,
	        "range $string_before .. $string_after has too many terms, limiting to $limit"
		    if $number <= $string_after_number
		    && ${$$vars_ref{constants}}{limits};

		# if we got rid of the first 3 strings in 
		# the array and didn't expand any back in,
		# we need to bump our middle string index

		$strings_index = 1 if $strings_index == 0;
	    }

	} # end if this string is a .. string
	
	# updated strings index
	$strings_index++;

    } # end looping through strings

    return ( $expanded );
}

#=======================================================================

# function to expand the macros etc in a list of text lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines, and
# modifies various structures in the flags/values hash.
#
# The following things happen in this function:
#   %constant definitions are stored and the lines swallowed
#   %define definitions are stored and the lines swallowed
#   invoked macros are expanded
#   %if %ifdef %ifndef conditionals are evaluated and applied

sub expand_lines
{
    my( $vars_ref, $lines_ref, $error_lines_ref, 
	$warning_lines_ref ) = @_;
    
    my $debug = ${$$vars_ref{constants}}{debug};
    print "$0: entering expand_lines\n" 
        if $debug;

    # tell of any skipping of expansions because 
    # of command line flags
    my $constants_ref = $$vars_ref{constants};
    print "$0: skipping macro expansions because of --nomacro flag\n"
        if $$constants_ref{nomacro};
    print "$0: skipping constant/define expansions because of --noconst flag\n"
        if $$constants_ref{noconst};
    print "$0: skipping conditionals expansions because of --nocond flag\n"
        if $$constants_ref{nocond};

    if ( $debug ) {
	print "$0: in expand_lines before expansions lines are:\n";
	my $index = 0;
	foreach my $line ( @$lines_ref ) {
	    printf "%s: line %04d: <%s>\n", $0, $index, $line;
	    $index++;
	}
    }

    # Loop through the list of lines, looping on each line
    # index while things keep getting changed at that index. 
    # All the conditional swallowing, and constant/define
    # definitions swallowing, and the macro expansions,
    # will happen at the line being examined and after so
    # we can use the index into the lines array to keep
    # track of out location and loop through the array.
    # Move on to the next line index when no changes are
    # made at the current index.

    my $line_index = 0;
    my $macro_depth = 0;
    my $macro_depth_limit = 100;
    while ( $line_index <= $#$lines_ref ) {
	my $changed = 0;
	my $line = $$lines_ref[$line_index];
	printf "%s: in expand_lines examining line %04d: <%s>\n", $0, $line_index, $line
	    if $debug;

	# the fields_witnhout_ruleset_comment function takes
	# a lot of cpu time if we call it more than we need to,
	# so we need to just call it once for every new line
	# being evaluated, if possible

	my @fields_without_comment = 
	    fields_without_ruleset_comment( $vars_ref, $line );

	# Swallow pfilter comment lines.  These are lines 
	# that have no content except for a pfilter comment
	# which starts with the % character. These do not
	# include blank lines. We also do not include
	# un-expanded constant/define invocations such
	# as %name% as part of the comment. It is safe to
        # to consume as many pfilter command only lines as
        # we find at this index.

	while ( $line =~ /[^\s]/ &&            # line is not blank
		! @fields_without_comment ) {
	    print "$0: in expand_lines at offset <$line_index> removing comment line <$line>\n"
		if $debug;
	    splice @$lines_ref, $line_index, 1; 
	    $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    @fields_without_comment = 
		fields_without_ruleset_comment( $vars_ref, $line );
	}

        # Expand any trailing conditional. We only do 
	# this for one conditional line and then jump
	# back to the loop start, because if a line is 
	# deleted at this index new lines it that are
	# moved up to this index might need variable 
	# expansions or other processing before
	# we try to do conditionals or macros there.

        if ( ! ${$$vars_ref{constants}}{nocond} &&
             expand_trailing_conditional_in_lines_at_offset( $vars_ref, 
							     $lines_ref, 
							     $line_index,
							     \@fields_without_comment,
							     $error_lines_ref,
							     $warning_lines_ref ) ) {
            $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since this line changed, start again at same line index
        }

	# Substitute the values for any referenced 
        # constants or defines on this line.
	# We do this before we look for new constant/
	# define definitions because we might have this:
	#        %define newvar %oldvar%
	
	if ( ! ${$$vars_ref{constants}}{noconst} &&
	     expand_constants_defines_in_lines_at_offset( $vars_ref,
							  $lines_ref,
							  $line_index,
							  $$vars_ref{constants},
							  $$vars_ref{defines},
							  $error_lines_ref,
							  $warning_lines_ref ) ) {
            $changed = 1;
	    if ( $debug ) {
		print "$0: in expand_lines original and substituted lines:\n" ;
		printf "%s: line %04d: <%s>\n", $0, $line_index, $line;
		printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
	    }
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    @fields_without_comment = 
		fields_without_ruleset_comment( $vars_ref, $line );
	}

        # Expand any conditional blocks that start at
        # this line index.  We only do this for one 
	# conditional block and then jump back to the
        # start of the loop, because if lines are 
	# deleted at this index new lines below them
	# that are moved up to this index might need
	# variable expansions or other processing before
	# we try to do conditionals or macros there.

        if ( ! ${$$vars_ref{constants}}{nocond} &&
             expand_conditional_block_in_lines_at_offset( $vars_ref, 
                                                          $lines_ref, 
                                                          $line_index,
                                                          \@fields_without_comment,
						          $error_lines_ref,
						          $warning_lines_ref ) ) {
            $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since this line changed, start again at same line index
        }

        # Swallow any constant/define definition lines at
	# this location, storing the new values.  We only do 
        # one constant/definition line and then jump back to
        # loop start, since we might have to do various types
        # of processing on the next line that is exposed.

	if ( parse_constant_define_from_line( $vars_ref,
                                              $line, 
                                              \@fields_without_comment,
                                              $error_lines_ref,
                                              $warning_lines_ref,
					      "" ) ) {
            splice @$lines_ref, $line_index, 1;
	    $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since this line changed, start again at same line index
        }

        # Expand any N .. M ranges.  We don't need to
        # jump back to rescan this line, since .. ranges
        # cannot intersect variable/constant names.
        # But if a .. range is expanded, we have to write
        # the modified fields back into the line being 
        # examined.

	if ( expand_dot_dot_ranges_in_strings( $vars_ref,
                                               \@fields_without_comment,
					       $error_lines_ref,
                                               $warning_lines_ref ) ) {
	    $$lines_ref[$line_index] = join( " ", 
                                             @fields_without_comment );
	}


	# Expand loops that start at this line index.
        # This is tricky because we can only do one loop 
        # expansion right now and and then we have to jump 
        # back to the loop start since we might have to do
        # various types of processing other than loop 
        # expansions on the new lines that are exposed.

	if ( ! ${$$vars_ref{constants}}{noloop} &&
             expand_loop_in_lines_at_offset( $vars_ref,
                                             $lines_ref, 
					     $line_index,
                                              \@fields_without_comment,
					     $error_lines_ref,
					     $warning_lines_ref ) ) {
	    $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since this line changed, start again at same line index
	}

	# Expand macros that start at this line index.
        # This is tricky because we have to guard against
        # infinite macro expansion loops at the same location, 
        # and we can only do one macro expansion right now and
        # and then we have to jump back to the loop start
        # since we might have to do various types of processing
        # other than macro expansions on the new lines that 
        # are exposed.

	if ( ! ${$$vars_ref{constants}}{nomacro} &&
	     $macro_depth < $macro_depth_limit &&
             expand_macro_in_lines_at_offset( $vars_ref,
					      $lines_ref, 
					      $line_index,
					      \@fields_without_comment,
					      $error_lines_ref,
					      $warning_lines_ref ) ) {
	    $macro_depth++;
	    push @$warning_lines_ref, "macro depth limit exceeded"
		if $macro_depth >= $macro_depth_limit;
	    $changed = 1;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since this line changed, start again at same line index
	}

        # swallow and process any %error or %warning metadirectvies
        # at this line index

        if ( expand_metadirective_in_lines_at_offset( $vars_ref, 
						      $lines_ref, 
						      $line_index,
						      \@fields_without_comment,
						      $error_lines_ref,
						      $warning_lines_ref ) ) {
            $changed = 1;
            printf "%s: in expand_lines absorbed metadirective line %04d: <%s>\n", $0, $line_index, $line
                if $debug;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    next; # since we swallowed a line, start again at same line index
        }

        # if there were no changes for this line index,
        # bump the index to the next line
        if ( ! $changed ) {
	    $line_index++;
	    last if $line_index > $#$lines_ref;
	    $line = $$lines_ref[$line_index];
	    $macro_depth = 0;
	}

    } # end looping on this line and incrementing through line indexes

    if ( $debug ) {
	print "$0: in expand_lines after expansion loops loops=$loop_count noexp=$no_changes_count lines:\n";
	my $index = 0;
	foreach my $line ( @$lines_ref ) {
	    printf "%s: line %04d: <%s>\n", $0, $index, $line;
	    $index++;
	}
    }
}

#=======================================================================

# function to expand a loop in a list of text lines at a given offset
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   offset into the lines array that has the loop invocation line
#   reference to array of non-blank fields excluding ruleset comments
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_loop_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $loop_start_index, 
	$fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};

#    if ( $debug ) {
#	my $string = ( $loop_start_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$loop_start_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_loop_in_lines_at_offset offset=$loop_start_index $string";
#    }

    # if we were passed an index past the array, done
    return 0 if $loop_start_index > $#$lines_ref;

    # find pointers to needed stuff an init other variables
    my $line_index = $loop_start_index;
    my $line = $$lines_ref[$line_index];

    # see if the line starts with a loop directive,
    # if so, we need to identify the complete loop
    # block and process it, skipping over nested loops,
    # if not, we're done
    return 0 
        if ! @$fields_without_comment_ref ||
        $$fields_without_comment_ref[0] ne "%loop";

    # find the loop variable and the loop values
    my $loop_variable_name;
    my @loop_values = @$fields_without_comment_ref;
    shift @loop_values;
    if ( @loop_values ) {
	$loop_variable_name = shift @loop_values;
	if ( $debug) {
	    my $string = join(",", @loop_values);
	    print "$0: in expand_loop_in_lines_at_offset loop variable name=<$loop_variable_name> values=($string)\n";
	}
	expand_dot_dot_ranges_in_strings( $vars_ref,
					  \@loop_values,
					  $error_lines_ref,
					  $warning_lines_ref );
    } else {
	push @$error_lines_ref, "loop no variable name to substitute";
	$loop_variable_name = "";
    }

    # find the end of the loop block making sure to
    # skip over nested loop blocks
    if ($debug ) {
	print "$0: in expand_loop_in_lines_at_offset block starting with:\n";
	printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
    }
    my $loop_level = 1;          # start at level 1 because 1 %loop found
    my $endloop_index = -1;      # -1 means no un-nested %endloop found yet
    my $search_index = $line_index + 1; # start at the line after the %loop
    # loop through lines until %endloop found or no more lines
    while ( $endloop_index == -1
	    && $search_index <= $#$lines_ref ) {
	my $search_line = $$lines_ref[$search_index];
	if ( $search_line =~ /^[\s]*\%loop(\s|$)/ ) {
	    $loop_level += 1;
	} elsif ( $search_line =~ /^\s*\%endloop(\s|$)/ ) {
	    $loop_level -= 1;
	    $endloop_index = $search_index
		if ! $loop_level;
	}
	# go to next line
	$search_index++;
    }
    print "$0: in expand_loop_in_lines_at_offset after searching level=$loop_level endloop=$endloop_index search=$search_index last=$#$lines_ref\n"
	if $debug;

    # if we cannot find the end of the loop block, note the error
    push @$error_lines_ref, "loop has no matching %endloop"
        if ( $endloop_index == -1 );

    # get rid of the line containing the initial %loop
    splice @$lines_ref, $line_index, 1;
    $endloop_index--
        if ( $endloop_index != -1 );
#    print "$0: in expand_loop_in_lines_at_offset after munching %loop line line_index=$line_index endloop=$endloop_index last=$#$lines_ref\n"
#	 if $debug;

    # if we found the end of the loop, ...
    if ( $endloop_index != -1 ) {
	
	# pull the loop lines out into a seperate array
	my $count = $endloop_index - $line_index;
	print "$0: in expand_loop_in_lines_at_offset line_index=$line_index endloop_index=$endloop_index count=$count\n"
	    if $debug;
	my @loop_lines = splice @$lines_ref, $line_index, $count;
	$endloop_index -= $count;
	
	# for each loop value, ...
	foreach my $loop_value ( @loop_values ) {
	    
	    # substitute the loop value into a new copy
	    # of the loop lines wherever the loop variable
	    # is referenced
	    my %substitution;
	    $substitution{$loop_variable_name} = $loop_value;
	    my %null;
	    my @new_lines = @loop_lines;
	    expand_constants_defines_in_lines( $vars_ref,
					       \@new_lines, 
					       \%null,
					       \%substitution,
					       $error_lines_ref,
					       $warning_lines_ref );
	    
	    # insert the variable substituted lines
	    splice @$lines_ref, $line_index, 0, @new_lines;
	    
	    # adjust the index of where we're at
	    $line_index += scalar @new_lines;
	    
	} # end for each loop value
	
	# get rid of %endloop line
	splice @$lines_ref, $line_index, 1;
	
    } # end if un-nested %endloop line was found

    return ( 1 );
}

#=======================================================================

# function to expand a macro in a list of text lines at a given offset
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   offset into the lines array that has the macro invocation line
#   reference to array of non-blank fields excluding ruleset comments
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_macro_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $macro_start_index,
	$fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};

#    if ( $debug ) {
#	my $string = ( $macro_start_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$macro_start_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_macro_in_lines_at_offset offset=$macro_start_index $string";
#    }

    # if we were passed an index past the array, done
    return 0 if ( $macro_start_index > $#$lines_ref );

    # find pointers to needed stuff an init other variables
    my $expanded = 0;
    my $line = $$lines_ref[$macro_start_index];
    my $macros_ref = $$vars_ref{macros};

    # find the line without any source/output comments
    # but we have to allow for constants/defines embedded
    # in the line so that a line like this:
    #         macro %define1% arg2 %define3% %comments
    # gets changed to this without the comments:        
    #         macro %define1% arg2 %define3%
    # we also want a list of all the non-comment fields
    my @wo_comment_fields = fields_without_ruleset_comment( $vars_ref, $line );

    # unless there is something on the line except comments,
    # and the first non-blank field is a %name% type field,
    # skip looking for macro expansion on it
    if ( @wo_comment_fields &&
	 $wo_comment_fields[0] =~ /^%[A-Za-z][A-Za-z0-9\_\-]*\%$/ ) {
	    
	# get rid of the leading and trailing % characters in the
	# macro name to find out what the supposed macro name is
	my $macro_name = shift @wo_comment_fields;
	$macro_name =~ s/^\%//;
	$macro_name =~ s/\%$//;
#        print "$0: in expand_macro_in_lines_at_offset possible macro_name=<$macro_name>\n" 
#	 if $debug;

	# there is a macro of that name, ...
	if ( exists $$macros_ref{$macro_name} ) {

            # get the leading white space from the line 
	    # so that we can add it to the expanded output
	    # at the beginning of lines
	    my $leading_space = $line;
	    $leading_space =~ s/\S.*$//;

	    # get a reference to the macro structure
	    my $macro_ref = $$macros_ref{$macro_name};
#	    print "$0: in expand_macro_in_lines_at_offset expanding macro $$macro_ref{name}\n"
#		if $debug;

	    # make a private copy of the macro text lines
	    my @new_lines = @{$$macro_ref{lines}};

	    # make a private copy of the supplied arguments
	    my @args_given = @wo_comment_fields;
#	    if ( $debug) {
#		my $string = join(",", @args_given);
#		print "$0: in expand_macro_in_lines_at_offset given argument list=($string)\n";
#	    }

	    # figure out the wanted number of macro paramters,
	    # and the supplied number of macro parameters
	    my $num_args_wanted = scalar @{$$macro_ref{arguments}};
	    my $num_args_given = scalar @args_given;

	    # build a hash of supplied arguments keyed off the macro
	    # substitition argument names, and warn of mismatches
	    my @args_list = ();
	    if ( $num_args_given < $num_args_wanted ) {
		push @$error_lines_ref,
		"macro $macro_name requires $num_args_wanted arguments, but is only supplied $num_args_given ( @args_given ), using blanks for missing arguments";
		@args_list = @args_given;
		while ( scalar @args_list < $num_args_wanted ) {
		    push @args_list, "";
		}
	    } elsif ( $num_args_given > $num_args_wanted ) {
		push @$error_lines_ref,
		"macro $macro_name requires $num_args_wanted arguments, but has been supplied $num_args_given ( @args_given ), ignoring unwanted arguments";
		@args_list = @args_given[0 .. ($num_args_wanted - 1)];;
	    } else {
		@args_list = @args_given;
	    }
#	    if ( $debug) {
#		my $string = join(",", @args_list);
#		print "$0: in expand_macro_in_lines_at_offset used argument list=($string)\n";
#	    }
	    my $macro_argument_names_ref = $$macro_ref{arguments};
	    my %args_hash;
	    foreach my $index ( 0 .. $#args_list ) {
		$args_hash{@$macro_argument_names_ref[$index]} =
		    $args_list[$index];
	    }
#	    print_hash( "", "in expand_macro_in_lines_at_offset argument hash", \%args_hash )
#		if $debug;

	    # make a private copy of the substituted macro lines
	    # and substitute the arguments into the new lines
	    # after inserting the leading white space that was
	    # on the macro invocation line at the beginning of
	    # each line
	    @new_lines = @{$$macro_ref{lines}};
	    foreach my $new_line ( @new_lines ) {
		$new_line =~ s/^/$leading_space/;
	    }
	    my %constants;
	    expand_constants_defines_in_lines( $vars_ref,
					       \@new_lines, 
					       \%constants,
					       \%args_hash,
					       $error_lines_ref,
					       $warning_lines_ref );

	    if ( $debug ) {
		print "$0: in expand_macro_in_lines_at_offset expanding macro=$macro_name in\n" ;
		printf "%s: line %04d: <%s>\n", $0, $macro_start_index, $line;
		my $debug_number = scalar @new_lines;
		print "$0: into these $debug_number lines:\n" ;
		my $debug_index = $macro_start_index;
		foreach my $debug_line ( @new_lines ) {
		    printf "%s: line %04d: <%s>\n", $0, $debug_index, $debug_line;
		    $debug_index++;
		}
	    }

	    # substitute the new macro lines in for the original 
	    # source line that invoked the macro
	    splice @$lines_ref, $macro_start_index, 1, @new_lines;

	    # set flag to show that we have done a macro expansion
	    $expanded = 1;
	}
    }

    return ( $expanded );
}

#=======================================================================

# function to expand/process/consume a metadirective in a list of text 
# lines at a given offset
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   offset into the lines array that has the starting %if? line
#   reference to array of non-blank fields excluding ruleset comments
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines
# Returned parameters are:
#   flag set to non-zero if any expansion occurred

sub expand_metadirective_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $line_index,
	$fields_without_comment_ref, 
	$error_lines_ref, $warning_lines_ref ) = @_;
#    my $debug = ${$$vars_ref{constants}}{debug};

#    if ( $debug ) {
#	my $string = ( $line_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$line_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_metadirectives_in_lines_at_offset offset=$line_index $string";
#    }

    # if we were passed an index past the array, done
    return 0 if ( $line_index > $#$lines_ref );

    # find pointers to needed stuff an init other variables
    my $expanded = 0;
    my $line = $$lines_ref[$line_index];

    # if nothing there after removing comments, done
    return 0 if ! @$fields_without_comment_ref;

    # try for the %error metadirective, if found,
    # insert the rest of the line after the metadirective
    # in the error lines list and swallow the line
    if ( $$fields_without_comment_ref[0] eq "%error" ) {
	my @fields = @$fields_without_comment_ref;
	shift @fields;                     # get rid of metadirective
	my $string = join( " ", @fields ); # get message
	push @$error_lines_ref, $string;   # insert the message
	splice @$lines_ref, $line_index, 1;# swallow the line
	$expanded = 1;                     # signify changes
    }

    # try for the %warning metadirective, if found,
    # insert the rest of the line after the metadirective
    # in the warning lines list and swallow the line
    if ( $$fields_without_comment_ref[0] eq "%warning" ) {
	my @fields = @$fields_without_comment_ref;
	shift @fields;                     # get rid of metadirective
	my $string = join( " ", @fields ); # get message
	push @$error_lines_ref, $string;   # insert the message
	splice @$lines_ref, $line_index, 1;# swallow the line
	$expanded = 1;                     # signify changes
    }

    return $expanded;
}

#=======================================================================

# function to expand/consume a single trailing conditional in a line at
# a given offset in a list of text lines
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   reference to array of text lines to expand
#   offset into the lines array that has the starting %if? line
#   reference to array of non-blank fields excluding ruleset comments
#   reference to error lines array
#   reference to warning lines array
# This function modifies the array of text lines, deleting the 
# line in question if the conditional evaluates as false, 
# or just deleting the "%if-- expression" at the end of the line
# if the conditional evaluates as true
# Returned parameters are:
#   This function returns non-zero if a conditional was processed
# Note that all referenced defines/constants in the conditional 
# arguments are substituted with their values before evaluation.

sub expand_trailing_conditional_in_lines_at_offset
{
    my( $vars_ref, $lines_ref, $conditional_start_index,
	$fields_without_comment_ref,
	$error_lines_ref, $warning_lines_ref ) = @_;
    my $debug = ${$$vars_ref{constants}}{debug};

#    if ( $debug ) {
#	my $string = ( $conditional_start_index <= $#$lines_ref ) ? 
#	    "line=<$$lines_ref[$conditional_start_index]>\n" : 
#		"last line index=$#$lines_ref\n";
#	print "$0: in expand_trailing_conditional_in_lines_at_offset offset=$conditional_start_index $string";
#    }

    return 0 if ( $conditional_start_index > $#$lines_ref );

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};

    my $line_index = $conditional_start_index;

    # If this line has a %if? conditional expression at the end 
    # of the line, evaluate it and either include the line (without
    # the conditional expression), or don't include the line at all.
    # Note that having onlye white space before the conditional does
    # not make it a trailing conditional, that makes it the beginning
    # of a conditional block which is caught and expanded by the
    # expand_conditional_block_in_lines_at_offset function.

    if ( scalar @$fields_without_comment_ref >= 2 &&
	 ! grep( /^%(if|ifdef|ifndef)$/, ( $$fields_without_comment_ref[0] ) ) &&
           grep( /^%(if|ifdef|ifndef)$/, @$fields_without_comment_ref ) ) {

	# find the portion of the line that is the conditional only
	my @conditional_fields = @$fields_without_comment_ref;
	shift @conditional_fields
	    while ( @conditional_fields &&
		    $conditional_fields[0] !~ /^%(if|ifdef|ifndef)$/ );
	my $conditional = join( " ", @conditional_fields );

	# see if the conditional has a syntax error and see
	# if it evaluates true
	my $syntax_error;
	my $evaluated;
	evaluate_conditional( $vars_ref,
			      $conditional,
			      $error_lines_ref,
			      $warning_lines_ref,
			      \$syntax_error,
			      \$evaluated );
	print "$0: in expand_trailing_conditionals_in_lines_at_offset after evaluating <$conditional> syntax_error=$syntax_error evaluated=$evaluated\n"
	    if $debug;
	
	# if the conditional had a syntax error or evaluated true
	# get rid of the conditional expression part of the line
	# and bump the index of the next line to look at
	if ( $syntax_error || $evaluated ) {
	    if ( $debug ) {
		print "$0: in expand_trailing_conditional_in_lines_at_offset evaluated true:\n" ;
		printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
	    }
	    $$lines_ref[$line_index] =~ s/\s+%(if|ifdef|ifndef)(\s+.*$|$)//;  # get rid of conditional
	    if ( $debug ) {
		print "$0: being retained as:\n" ;
		printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
	    }
	}
	
	# if there was no syntax error and the conditional
	# evaluated false swallow the line
	else {
	    if ( $debug ) {
		print "$0: in expand_trailing_conditional_in_lines_at_offset swallowing evaluated false:\n" ;
		printf "%s: line %04d: <%s>\n", $0, $line_index, $$lines_ref[$line_index];
	    }
	    splice @$lines_ref, $line_index, 1;
	}
	
	return 1;

    # this line didn't have a conditional expression on the end of
    # the line, so just return a zero to show that it wasn't modified

    } else {

	return 0;

    }
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;
