# Pfilter/Rulesets.pl
#
# common library functions for the ruleset parts of pfilter

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# function to parse a single ruleset file
# sub parse_ruleset_file( $vars_ref, $ruleset_file, 
# 			  $error_lines_ref, $warning_lines_ref )

# function to parse all of the ruleset files
# sub parse_ruleset_files( $vars_ref, $ruleset_files_directory, 
#                          $required_file )

#=======================================================================

# function to parse a single ruleset file
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   name of the ruleset file to parse
#   reference to error lines array
#   reference to warning lines array
# This function modifies the hash of all flags/values, with new/modified:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
#     macros -> reference to %MACRO names/references hash
# Returned parameters are:
#   flag which is non-zero if file successfully parsed

sub parse_ruleset_file
{
    my( $vars_ref, $ruleset_file, 
	$error_lines_ref, $warning_lines_ref ) = @_;

    my $debug = ${$$vars_ref{constants}}{debug};     # save time
    print "$0: in parse_ruleset_file ruleset_file=$ruleset_file\n"
	if ( $debug );

    # find pointers to needed stuff
    my $constants_ref = $$vars_ref{constants};
    my $defines_ref = $$vars_ref{defines};
    my $macros_ref = $$vars_ref{macros};

    # open the ruleset file and read it in
    if ( ! open( RULESET, $ruleset_file ) ) {
	print "$0: cannot open ruleset file $ruleset_file\n";
	return 0;
    }
    my @lines = <RULESET>;
    close(RULESET);
    chomp @lines;
#    if ( $debug ) {
#	my $numlines = @lines;
#	print "$0: read $numlines lines from $ruleset_file\n";
#    }

    # loop through the input lines
    my $line_number = 0;
    my $macro_name;
    my $macro_ref;
    my $macro_level = 0;
    foreach my $line ( @lines ) {
	$line_number++;
	my $where = "on line $line_number in $ruleset_file";

	# find the line without any comments to help out later logic
	my $wo_comments = $line;              # get rid of comments ...
	$wo_comments =~ s/^\%\s.*$//;         # begin-line % space anything
	$wo_comments =~ s/\s+\%[^\%]*\s.*$//; # space(s) % non-%(s) space anything
	$wo_comments =~ s/\s+\%[^\%]*$//;     # space(s) % non-%(s) end-of-line
	$wo_comments =~ s/^\%$//;             # begin-line % end-of-line

	# split the non-comment portion of the line into non-space fields
	my @fields = split(' ',$wo_comments);

	# if the line is a %macro start of macro definition,
	# increase the macro level count if we are already in
	# the middle of a macro, or start a new macro
        if ( @fields && $fields[0] eq "%macro" ) {
	    if ( ! $macro_level ) {
		# if no macro name, error
		if ( @fields < 2 ) {
		    push @$error_lines_ref, 
		    "no macro name or arguments $where\n";
		} else {
		    # get the name of the macro
		    shift @fields;
		    $macro_name = shift @fields;
		    $macro_name =~ s/^\%//;
#		    print "$0: starting definition of macro $macro_name $where\n"
#			if $debug;
		    # see if this macro already exists and if so,
		    # warn about redefining the macro
		    push @$warning_lines_ref, 
		    "macro $macro_name being redefined $where\n"
			if exists $$macros_ref{$macro_name};
		    # initialize a new macro structure and point 
		    # macros hash entry for it's name to it
		    my %macro;
		    $macro_ref = \%macro;
		    $$macros_ref{$macro_name} = $macro_ref;
		    $$macro_ref{name} = $macro_name;
		    my @macro_arguments = @fields;
		    my @macro_lines = ();
		    $$macro_ref{arguments} = \@macro_arguments;
		    $$macro_ref{lines} = \@macro_lines;
		    $$macro_ref{defined_file_name} = $ruleset_file;
		    $$macro_ref{defined_line_number} = $line_number;
#		    print_hash("", "initialized macro $macro_name", $macro_ref)
#			if $debug;
		}
	    }
	    $macro_level = $macro_level + 1;

	# if the line is a %endmacro end of macro definition,
	# decrease the macro level count if we are in nested
        # macro definitions, or close the current macro if not
	} elsif ( @fields && $fields[0] eq "%endmacro" ) {
	    if ( $macro_level ) {
		$macro_level = $macro_level - 1;
		if ( ! $macro_level ) {
#		    print "$0: ending definition of macro $macro_name $where\n"
#			if $debug;
#		    print_hash("", "finished macro $macro_name", $macro_ref)
#			if $debug;
		}
	    } else {
		push @$error_lines_ref, "%endmacro without %macro $where";
	    }

	# if we are in the middle of a macro definition,
	# store this line in the macro definition
	} elsif ( $macro_level ) {
	    push @{$$macro_ref{lines}}, $line;
#	    #print "$0: macro $macro_name now has " . scalar @{$$macro_ref{lines}} . " lines\n"
#                #if $debug;

        # if this line is a %constant store the constant
	} elsif ( @fields && $fields[0] eq "%constant" ) {
            # if no value being assigned, error
	    if ( @fields < 3 ) {
		push @$error_lines_ref, 
		"no value for %constant $fields[1] $where\n";
	    # if the name is not a legal name, error
            } elsif ( $fields[1] !~ /[A-Za-z\_][A-Za-z0-9\_\-]*/ ) {
		push @$error_lines_ref, 
		"illegal %constant name <$fields[1]> $where\n";
	    # if a constant with this name is already defined,
	    # cannot redefine it
	    } elsif ( exists $$constants_ref{$fields[1]} ) {
		push @$error_lines_ref, 
		"illegal attempt to redefine %constant $fields[1] $where";
	    } else {
		# if a %define has been done already, 
		# delete the define and warn of it 
		# before storing the new %constant
		if ( exists $$defines_ref{$fields[1]} ) {
		    push @$warning_lines_ref,
		    "%define $fields[1] being redefined as %constant $fields[1] $where";
		    delete $hash{$name};
                }
                # store the new %constant by taking the non-comment
                # part of the input line and getting rid of the 
                # first two non-white-space/white-space groups
                my $value = $wo_comments;
                $value =~ s/^[^\s]+\s+[^\s]+\s+//;
		$$constants_ref{$fields[1]} = $value;			       
            }

        # if this line is a %define store the define
	} elsif ( @fields && $fields[0] eq "%define" ) {
            # if no value being assigned, error
	    if ( @fields < 3 ) {
		push @$error_lines_ref, 
		"no value for %define $fields[1] $where\n";
	    # if the name is not a legal name, error
            } elsif ( $fields[1] !~ /[A-Za-z\_][A-Za-z0-9\_\-]*/ ) {
		push @$error_lines_ref, 
		"illegal %define name <$fields[1]> $where\n";
	    # if a constant with this name is already defined,
	    # cannot redefine it
	    } elsif ( exists $$constants_ref{$fields[1]} ) {
		push @$error_lines_ref, 
		"illegal attempt to redefine %constant $fields[1] as %define $fields[1] $where";
	    } else {
		# if a %define has been done already, 
		# delete the define and warn of it 
		# before storing the new %define
		if ( exists $$defines_ref{$fields[1]} ) {
		    push @$warning_lines_ref,
		    "%define $fields[1] being redefined as %constant $fields[1] $where";
		    delete $hash{$name};
                }
                # store the new %define by taking the non-comment
                # part of the input line and getting rid of the 
                # first two non-white-space/white-space groups
                my $value = $wo_comments;
                $value =~ s/^[^\s]+\s+[^\s]+\s+//;
		$$defines_ref{$fields[1]} = $value;			       
            }

	# if this line has stuff other than comments, issue warning
	} elsif ( $wo_comments ne "" ) {
	    push @$warning_lines_ref, "unknown rulesets text $where";
	}

    # end of loop for each line in this rulesets file
    }

    # if a macro is currently being defined as we finish
    # the rulesets file, output an error message
    push @$error_lines_ref, "macro $$macro_ref{name} definition that was started on line $$macro_ref{defined_line_number} of ruleset file $$macro_ref{defined_file_name} not finished\n"
	if $macro_level;

}

#=======================================================================

# function to parse all of the ruleset files
#
# Passed parameters are:
#   reference to hash of all flags/values of interest:
#   directory name that has all the ruleset files
#   name of required file that is processed before other files
# This function modifies the hash of all flags/values, with new/modified:
#     constants -> reference to %CONSTANT names/values hash
#     defines -> reference to %DEFINE names/values hash
#     macros -> reference to %MACRO names/references hash
# Returned parameters are:
#   reference to an array of error messages lines
#   reference to an array of warning message lines
#
# For example, after calling the function as follows:

sub parse_ruleset_files
{
    my( $vars_ref, $ruleset_files_directory, $required_file ) = @_;

    $debug = ${$$vars_ref{constants}}{debug};     # save time
#    print "$0: in parse_ruleset_filess ruleset_files_directory=$ruleset_files_directory required_file=$required_file\n"
#	if ( $debug );

    # find pointers to needed stuff or make new structures
    my $constants_ref;
    if ( ! exists $$vars_ref{constants} ) {
	my %constants;
        $$vars_ref{constants} = \%constants;
    }
    $constants_ref = $$vars_ref{constants};
    my $defines_ref;
    if ( ! exists $$vars_ref{defines} ) {
	my %defines;
        $$vars_ref{defines} = \%defines;
    }
    $defines_ref = $$vars_ref{defines};
    my $macros_ref;
    if ( ! exists $$vars_ref{macros} ) {
	my %macros;
        $$vars_ref{macros} = \%macros;
    }
    $macros_ref = $$vars_ref{macros};
    my @error_lines;
    my @warning_lines;

    # get a listing of the rulesets directory
    opendir( RULESETS, $ruleset_files_directory ) ||
	die "$0: cannot open rulesets directory $ruleset_files_directory\n";
    my @directory_entries = readdir( RULESETS );
    closedir RULESETS;

    # find the files not ending with ~ in the rulesets directory
    # (this gets rid of emacs backup files in case we're debugging)
    my @ruleset_files;
    foreach my $filename ( grep(!/~$/, (@directory_entries) ) ) {
	chomp $filename;
	push @ruleset_files, "$ruleset_files_directory/$filename"
	    if -f "$ruleset_files_directory/$filename";
    }
#    print "$0: ruleset_files=@ruleset_files\n" if $debug;

    # make sure the one required pfilter rulesets file is in the list
    die "$0: cannot find required rulesets file $ruleset_files_directory/$required_file\n"
	if ( $required_file ne "" &&
	     ! grep /^$ruleset_files_directory\/$required_file$/, @ruleset_files );

    # re-order list so that the one required pfilter rulesets file is first
    # so that users can override the default rulesets in any other file
    my @ordered_ruleset_files;
    if ( $required_file ne "" ) {
	@ordered_ruleset_files = 
	    grep ( !/^$ruleset_files_directory\/$required_file$/, @ruleset_files );
	unshift @ordered_ruleset_files, "$ruleset_files_directory\/$required_file";
        print "$0: ordered_ruleset_files=@ordered_ruleset_files\n" 
            if $debug;
    } else {
	@ordered_ruleset_files = @ruleset_files;
    }

    # parse each of the ruleset files
    foreach my $ruleset_file ( @ordered_ruleset_files ) {
	parse_ruleset_file( $vars_ref, $ruleset_file, 
			    \@error_lines, \@warning_lines );
    }

#    print "error_lines=@error_lines warning_lines=@warning_lines\n";
    return ( \@error_lines, \@warning_lines );
}

#=======================================================================
#
# return value of 1 so that programs can call this library with "require"

1;
