# Pfilter/Subs.pl
#
# misc library functions for pfilter

# Copyright 2003 Neil Gorsuch
#
# This file is part of pfilter.
#
# pfilter is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# pfilter is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

#=======================================================================
# functions in this file:

# deep_copy( $item )

# usage()

# valid_addresses( $vars_ref, $addresses )

# valid_bcast( $vars_ref, $bcast)

# valid_destination( $vars_ref, $destination )

# valid_destinations( $vars_ref, $destinations )

# valid_internal_address( $vars_ref, $internal_address )

# valid_port_range( $vars_ref, $protocol, $range )

# valid_ports( $vars_ref, $protocol, $ports )

# valid_source( $vars_ref, $source )

# valid_sources( $vars_ref, $sources )

#=======================================================================

sub deep_copy {
    my $this = shift;
    if ( not ref $this ) {
	$this;
    } elsif ( ref %this eq "ARRAY" ) {
	[map deep_copy($_), @$this];
    }
#    elsif ( ref %this eq "HASH" ) {
#	+{map {$_ => deep_copy($this->{$_}) } keys $$this};
#    }
}

#=======================================================================

# note secret debug flags --nocond --noconst --nomacro not shown
sub usage {
    my ( $vars_ref ) = @_;
    print <<ENDIT;
Usage: $0 [ switches ... ] [start|stop|restart|status]
  --config=FILE     specify configuration file location
                    - means use STDIN
  --debug or -d     turn on debug output
ENDIT
    print <<ENDIT if ${$$vars_ref{constants}}{debug};
  --dumpconf        DEBUG: dumps the parsed config file hash
  --dumpinterfaces  DEBUG: dumps the final interfaces data structure
  --dumpservices    DEBUG: dumps the parsed ruleset defined services
ENDIT
    print <<ENDIT;
  --help or -h      print this help, then exit
  --ipchains        force usage of ipchains
  --iptables        force usage if iptables
  --logging         sets logging level to limited
  --logging=full    sets logging level to full
  --logging=limited sets logging level to limited
  --logging=none    sets logging level to none
ENDIT
    print <<ENDIT if ${$$vars_ref{constants}}{debug};
  --nocond          DEBUG: do not expand conditionals
  --noconst         DEBUG: do not expand constants or defines
ENDIT
    print <<ENDIT;
  --nodebug         turn off debug output (default)
  --noexec          generate but don't execute output file
  --nolimits        turns off looping/range/expansion limits
  --nologging       sets logging level to none
ENDIT
    print <<ENDIT if ${$$vars_ref{constants}}{debug};
  --noloop          DEBUG: do not expand loops
  --nomacro         DEBUG: do not expand macros
  --noprefix        DEBUG: do not insert pfilter-prefix macro
  --nosuffix        DEBUG: do not insert pfilter-suffix macro
ENDIT
    print <<ENDIT;
  --nooptpath       don't optimize source/destination paths
  --noout           don't generate or execute output file
  --nowarnings      disable warnings (and Perl syntax warnings)
  --output=FILE     specify output rules file location
                    - means use STDOUT and forces --noexec
ENDIT
    print <<ENDIT if ${$$vars_ref{constants}}{debug};
  --source=FILE     DEBUG: write out intermediate source file
                           - uses STDOUT and forces --noout
ENDIT
    print <<ENDIT;
  --verbose or -v   verbose output
  --version         print version number, then exit
  --warnings or -w  enable warnings (and Perl syntax warnings)

please report bugs to ngorsuch\@ncsa.uiuc.edu
ENDIT
}

#=======================================================================

sub valid_addresses {
    my ( $vars_ref, $addresses ) = @_;
    print "$0: in valid_addresses addresses=$addresses\n" if ${$$vars_ref{constants}}{debug};
    my @addresses = split /\,/, $addresses;
    return () unless @addresses;
    foreach my $address (@addresses) {
	my @resolved_ips = ();
	return () unless valid_address( $vars_ref, $address, \@resolved_ips );
    }
    print "$0: in valid_addresses returning @addresses\n" if ${$$vars_ref{constants}}{debug};
    return @addresses;
}

#=======================================================================

sub valid_bcast {
    my ($vars_ref, $bcast) = @_;
#    print "$0: in valid_bcast bcast=$bcast\n" if ${$$vars_ref{constants}}{debug};
    my @fields = $bcast =~ /^(\d+)\.(\d+)\.(\d+)\.(\d+)$/;
    return 0 if @fields != 4;
    if (($fields[3] & 3) != 3) {
	print "$0: in valid_bcast bad last ip chunk $fields[3]\n";
	return 0;
    }
    foreach my $num ( @fields ) {
	next if $num <= 255;
	print "$0: in valid_bcast invalid ip chunk $num\n";
	return 0;
    }
    return 1;
}

#=======================================================================

sub valid_destination {
    my ($vars_ref, $destination ) = @_;
#    print "$0: in valid_destination destination=$destination\n" if ${$$vars_ref{constants}}{debug};
    my @resolved_ips;
    return valid_address($vars_ref, $destination, \@resolved_ips);
}

#=======================================================================

sub valid_destinations {
    my ($vars_ref, $destinations) = @_;
#    print "$0: in valid_destinations destinations=$destinations\n" if ${$$vars_ref{constants}}{debug};
    my @resolved_ips;
    return valid_addresses($vars_ref, $destinations, \@resolved_ips);
}

#=======================================================================

sub valid_internal_address {
    my ($vars_ref, $internal_address) = @_;
#    print "$0: in valid_internal_address internal_address=$internal_address\n" if ${$$vars_ref{constants}}{debug};
    my @resolved_ips;
    return valid_address($vars_ref, $internal_address, \@resolved_ips);
}

#=======================================================================

sub valid_port_range {
    my ($vars_ref, $protocol, $range) = @_;
#    print "$0: in valid_port_range protocol=$protocol range=$range\n" if ${$$vars_ref{constants}}{debug};
    $range =~ s/\s+//g;
    my @fields = split /(-|\.\.|:)/, $range;
    if (@fields == 1) {
	my ($low) = $fields[0];
	return 1 if valid_port($vars_ref, $protocol, $low);
    } elsif (@fields == 2) {
	my ($low, $high) = @fields;
	return 1 if valid_port($vars_ref, $protocol, $low) &&
	    valid_port($protocol, $high) &&
		$high >= $low;
    }
    return 0;
}

#=======================================================================

sub valid_ports {
    my ($vars_ref, $protocol, $ports) = @_;
#    print "$0: in valid_ports protocol=$protocol ports=$ports\n" if ${$$vars_ref{constants}}{debug};
    my @ranges = split /,/, $ports;
    return () unless @ranges;
    foreach my $range (@ranges) {
	return () unless valid_port_range($vars_ref, $protocol, $range);
    }
#    print "$0: in valid_ports returning ranges=@ranges\n" if ${$$vars_ref{constants}}{debug};
    return @ranges;
}

#=======================================================================

sub valid_source {
    my ($vars_ref, $source) = @_;
#    print "$0: in valid_source source=$source\n" if ${$$vars_ref{constants}}{debug};
    my @resolved_ips;
    return valid_address($vars_ref, $source, \@resolved_ips);
}

#=======================================================================

sub valid_sources {
    my ($vars_ref, $sources) = @_;
#    print "$0: in valid_sources sources=$sources\n" if ${$$vars_ref{constants}}{debug};
#    print "$0: in valid_sources returning ",valid_addresses($vars_ref, $sources),
#    "\n" if ${$$vars_ref{constants}}{debug};
    return valid_addresses($vars_ref, $sources);
}

1;
